import inspect

from . import datasets
from .datasets import CLASSIFICATION, REGRESSION

all_datasets = [cls for _, cls in inspect.getmembers(datasets)
                if inspect.isclass(cls)
                and issubclass(cls, datasets.Dataset)
                and cls != datasets.Dataset
                and cls != datasets.ClassificationDataset
                and cls != datasets.RegressionDataset]


def regression_datasets():
    return [ds for ds in all_datasets if ds.type_ == REGRESSION]


def classification_datasets():
    return [ds for ds in all_datasets if ds.type_ == CLASSIFICATION]



import os
import abc
from copy import deepcopy

import numpy as np
import pandas as pd
from unlzw import unlzw

from ucimlr.helpers import (download_file, download_unzip, one_hot_encode_df_, xy_split, label_encode_df_,
                            clean_na_, normalize_df_, split_normalize_sequence, split_df, get_split, split_df_on_column,
                            split_classification_df)
from ucimlr.constants import TRAIN, VALIDATION
from ucimlr.constants import REGRESSION, CLASSIFICATION


class Dataset(abc.ABC):
    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, i):
        return self.x[i], self.y[i]

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def num_features(self):
        return self.x.shape[1]


class ClassificationDataset(Dataset):
    type_ = CLASSIFICATION  # Is this necessary?

    @property
    def num_classes(self):
        raise NotImplementedError('A classification dataset needs to have attribute `num_classes`.')


class RegressionDataset(Dataset):
    type_ = REGRESSION  # Is this necessary?

    @property
    def num_targets(self):
        return self.y.shape[1]


class Abalone(RegressionDataset):
    """
    Link to the dataset [description](http://archive.ics.uci.edu/ml/datasets/Abalone).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'data.csv'
        file_path = os.path.join(dataset_path, filename)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/abalone/abalone.data'
        download_file(url, dataset_path, filename)
        df = pd.read_csv(file_path, header=None)
        y_columns = df.columns[-1:]
        one_hot_encode_df_(df)
        df_test, df_train, df_valid = split_df(df, [0.2, 0.8 - 0.8 * validation_size, 0.8 * validation_size])
        normalize_df_(df_train, other_dfs=[df_valid, df_test])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)


class Adult(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 2

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename_train = 'data_train.csv'
        filename_test = 'data_test.csv'
        file_path_train = os.path.join(dataset_path, filename_train)
        file_path_test = os.path.join(dataset_path, filename_test)
        url_train = 'http://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data'
        url_test = 'http://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test'
        download_file(url_train, dataset_path, filename_train)
        download_file(url_test, dataset_path, filename_test)

        df_train_valid = pd.read_csv(file_path_train, header=None, skiprows=0)
        df_test = pd.read_csv(file_path_test, header=None, skiprows=1)

        # Trailing period in test file
        df_test[14] = df_test[14].str.rstrip('.')

        df_test.index += len(df_train_valid)
        df = pd.concat([df_train_valid, df_test])
        y_columns = df.columns[-1:]
        one_hot_encode_df_(df, skip_columns=y_columns)
        label_encode_df_(df, y_columns[0])
        df_train_valid, df_test = (df.loc[df_train_valid.index], df.loc[df_test.index])
        df_train, df_valid = split_df(df_train_valid, [1 - validation_size, validation_size])
        normalize_df_(df_train, other_dfs=[df_valid, df_test], skip_column=y_columns[0])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)
        self.y = self.y[:, 0]  # Flatten for classification


class AirFoil(RegressionDataset):
    """
    Link to the dataset [description](http://archive.ics.uci.edu/ml/datasets/Airfoil+Self-Noise).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'airfoil_self_noise.dat'
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00291/airfoil_self_noise.dat'
        download_file(url, dataset_path, filename)
        file_path = os.path.join(dataset_path, filename)

        df = pd.read_csv(file_path, sep='\t', header=None)
        y_columns = [5]
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class AirQuality(RegressionDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'AirQualityUCI.csv'
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00360/AirQualityUCI.zip'
        download_unzip(url, dataset_path)
        file_path = os.path.join(dataset_path, filename)

        df = pd.read_csv(file_path, sep=';', parse_dates=[0, 1])
        df.dropna(axis=0, how='all', inplace=True)
        df.dropna(axis=1, how='all', inplace=True)

        df.Date = (df.Date - df.Date.min()).astype('timedelta64[D]')  # Days as int
        df.Time = df.Time.apply(lambda x: int(x.split('.')[0]))  # Hours as int
        df['C6H6(GT)'] = df['C6H6(GT)'].apply(lambda x: float(x.replace(',', '.')))  # Target as float

        # Some floats are given with ',' instead of '.'
        df = df.applymap(lambda x: float(x.replace(',', '.')) if type(x) is str else x)  # Target as float

        df = df[df['C6H6(GT)'] != -200]  # Drop all rows with missing target values
        df.loc[df['CO(GT)'] == -200, 'CO(GT)'] = -10  # -200 means missing value, shifting this to be closer to
        # the other values for this column

        y_columns = ['C6H6(GT)']
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class APSFailure(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 2

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        file_name_train = 'train.csv'
        file_name_test = 'test.csv'
        dataset_path = os.path.join(root, self.name)
        url_train = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00421/aps_failure_training_set.csv'
        url_test = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00421/aps_failure_test_set.csv'
        download_file(url_train, dataset_path, file_name_train)
        download_file(url_test, dataset_path, file_name_test)
        file_path_train = os.path.join(dataset_path, file_name_train)
        file_path_test = os.path.join(dataset_path, file_name_train)
        df_train_valid = pd.read_csv(file_path_train, skiprows=20, na_values='na')
        df_test = pd.read_csv(file_path_test, skiprows=20, na_values='na')

        # TODO This is risky business since test and train might be cleaned to have different columns
        clean_na_(df_train_valid)
        clean_na_(df_test)
        if not (df_train_valid.columns == df_test.columns).all():
            raise Exception('Cleaning lead to different set of columns for train/test')

        y_columns = ['class']
        label_encode_df_([df_train_valid, df_test], y_columns[0])
        df_train, df_valid = split_df(df_train_valid, [1 - validation_size, validation_size])
        normalize_df_(df_train, other_dfs=[df_valid, df_test], skip_column=y_columns[0])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)
        self.y = self.y[:, 0]


class Avila(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 12

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00459/avila.zip'
        download_unzip(url, dataset_path)
        file_path_train = os.path.join(dataset_path, 'avila', 'avila-tr.txt')
        file_path_test = os.path.join(dataset_path, 'avila', 'avila-ts.txt')
        df_train_valid = pd.read_csv(file_path_train, header=None)
        df_test = pd.read_csv(file_path_test, header=None)
        y_columns = [10]
        label_encode_df_([df_train_valid, df_test], y_columns[0])  # Assumes encoding will be identical for train/test
        df_train, df_valid = split_classification_df(df_train_valid,
                                                     [1 - validation_size, validation_size],
                                                     y_columns[0])
        normalize_df_(df_train, other_dfs=[df_valid, df_test], skip_column=y_columns[0])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)
        self.y = self.y[:, 0]


class BankMarketing(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 2

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00222/bank-additional.zip'
        download_unzip(url, dataset_path)
        file_path = os.path.join(dataset_path, 'bank-additional', 'bank-additional-full.csv')
        df = pd.read_csv(file_path, sep=';')
        y_columns = ['y']
        one_hot_encode_df_(df, skip_columns=y_columns)
        label_encode_df_(df, y_columns[0])
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class BlogFeedback(RegressionDataset):
    """
    Link to the dataset [description](http://archive.ics.uci.edu/ml/datasets/BlogFeedback).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        file_name = 'blogData_train.csv'
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00304/BlogFeedback.zip'
        download_unzip(url, dataset_path)

        # Iterate all test csv and concatenate to one DataFrame
        test_dfs = []
        for fn in os.listdir(dataset_path):
            if 'blogData_test' not in fn:
                continue
            file_path = os.path.join(dataset_path, fn)
            test_dfs.append(pd.read_csv(file_path, header=None))
        df_test = pd.concat(test_dfs)

        file_path = os.path.join(dataset_path, file_name)
        df_train_valid = pd.read_csv(file_path, header=None)
        y_columns = [280]
        df_train_valid[y_columns[0]] = np.log(df_train_valid[y_columns[0]] + 0.01)
        df_test[y_columns[0]] = np.log(df_test[y_columns[0]] + 0.01)

        page_columns = list(range(50))
        for i, (_, df_group) in enumerate(df_train_valid.groupby(page_columns)):
            df_train_valid.loc[df_group.index, 'page_id'] = i
        df_train, df_valid = split_df_on_column(df_train_valid, [1 - validation_size, validation_size], 'page_id')
        df_train.drop(columns='page_id', inplace=True)
        df_valid.drop(columns='page_id', inplace=True)

        normalize_df_(df_train, other_dfs=[df_valid, df_test])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)


class CardDefault(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 2

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'data.xls'
        file_path = os.path.join(dataset_path, filename)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00350/' \
              'default%20of%20credit%20card%20clients.xls'
        download_file(url, dataset_path, filename)
        df = pd.read_excel(file_path, skiprows=1, index_col='ID')
        y_columns = ['default payment next month']
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class CTSlices(RegressionDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00206/slice_localization_data.zip'
        download_unzip(url, dataset_path)
        file_name = 'slice_localization_data.csv'
        file_path = os.path.join(dataset_path, file_name)
        df = pd.read_csv(file_path)
        # No patient should be in both train and test set
        df_train_valid = deepcopy(df.loc[df.patientId < 80, :])  # Pandas complains if it is a view
        df_test = deepcopy(df.loc[df.patientId >= 80, :])        # - " -
        df_train, df_valid = split_df_on_column(df_train_valid, [1 - validation_size, validation_size], 'patientId')
        y_columns = ['reference']
        normalize_df_(df_train, other_dfs=[df_valid, df_test])
        df_res = get_split(df_train, df_valid, df_test, split)
        df_res = df_res.drop(columns='patientId')
        self.x, self.y = xy_split(df_res, y_columns)


class FacebookComments(RegressionDataset):
    """
    Predict the number of likes on posts from a collection of Facebook pages.
    Every page has multiple posts, making the number of pages less than the samples
    in the dataset (each sample is one post).

    # Note
    The provided test split has a relatively large discrepancy in terms
    of distributions of the features and targets. Training and validation splits are
    also made to ensure that the same page is not in both splits. This makes the distributions
    of features in training and validation splits vary to a relatively large extent, possible
    because the number of pages are not that many, while the features are many.

    Link to the dataset [description](http://archive.ics.uci.edu/ml/datasets/Facebook+Comment+Volume+Dataset).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00363/Dataset.zip'
        download_unzip(url, dataset_path)
        dataset_path = os.path.join(dataset_path, 'Dataset')

        # The 5th variant has the most data
        train_path = os.path.join(dataset_path, 'Training', 'Features_Variant_5.csv')
        test_path = os.path.join(dataset_path, 'Testing', 'Features_TestSet.csv')
        df_train_valid = pd.read_csv(train_path, header=None)
        df_test = pd.read_csv(test_path, header=None)
        y_columns = df_train_valid.columns[-1:]

        # Page ID is not included, but can be derived. Page IDs can not be
        # in both training and validation sets
        page_columns = list(range(29))
        for i, (_, df_group) in enumerate(df_train_valid.groupby(page_columns)):
            df_train_valid.loc[df_group.index, 'page_id'] = i
        df_train, df_valid = split_df_on_column(df_train_valid, [1 - validation_size, validation_size], 'page_id')
        df_train.drop(columns='page_id', inplace=True)
        df_valid.drop(columns='page_id', inplace=True)

        normalize_df_(df_train, other_dfs=[df_valid, df_test])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)


class Landsat(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 6

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        file_name_train = 'train.csv'
        file_name_test = 'test.csv'
        url_train = 'http://archive.ics.uci.edu/ml/machine-learning-databases/statlog/satimage/sat.trn'
        url_test = 'http://archive.ics.uci.edu/ml/machine-learning-databases/statlog/satimage/sat.tst'
        download_file(url_train, dataset_path, file_name_train)
        download_file(url_test, dataset_path, file_name_test)
        file_path_train = os.path.join(dataset_path, file_name_train)
        file_path_test = os.path.join(dataset_path, file_name_test)
        df_train_valid = pd.read_csv(file_path_train, sep=' ', header=None)
        df_test = pd.read_csv(file_path_test, sep=' ', header=None)
        df_test.index += len(df_train_valid)
        df = pd.concat([df_train_valid, df_test])
        y_columns = [36]
        label_encode_df_(df, y_columns[0])
        df_train_valid = df.loc[df_train_valid.index, :]
        df_test = df.loc[df_test.index, :]
        df_train, df_valid = split_classification_df(df_train_valid, [1 - validation_size, validation_size], 36)
        normalize_df_(df_train, other_dfs=[df_valid, df_test], skip_column=y_columns[0])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)
        self.y = self.y[:, 0]


class LetterRecognition(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 26

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        file_name = 'data.csv'
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/letter-recognition/letter-recognition.data'
        download_file(url, dataset_path, file_name)
        file_path = os.path.join(dataset_path, file_name)
        df = pd.read_csv(file_path, header=None)
        y_columns = [0]
        label_encode_df_(df, y_columns[0])
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class MagicGamma(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 2

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        file_name = 'data.csv'
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/magic/magic04.data'
        download_file(url, dataset_path, file_name)
        file_path = os.path.join(dataset_path, file_name)
        df = pd.read_csv(file_path, header=None)
        y_columns = [10]
        label_encode_df_(df, y_columns[0])
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class OnlineNews(RegressionDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00332/OnlineNewsPopularity.zip'
        download_unzip(url, dataset_path)
        file_path = os.path.join(dataset_path, 'OnlineNewsPopularity', 'OnlineNewsPopularity.csv')
        df = pd.read_csv(file_path, )
        df.drop(columns=['url', ' timedelta'], inplace=True)
        y_columns = [' shares']
        df[y_columns[0]] = np.log(df[y_columns[0]])
        self.x, self. y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class Parkinson(RegressionDataset):
    """
    Link to the dataset [description](http://archive.ics.uci.edu/ml/datasets/parkinsons).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'data.csv'
        file_path: str = os.path.join(dataset_path, filename)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/' \
              'parkinsons/telemonitoring/parkinsons_updrs.data'
        download_file(url, dataset_path, filename)
        df = pd.read_csv(file_path)
        y_columns = ['motor_UPDRS', 'total_UPDRS']

        df_train_valid = df[df['subject#'] <= 30]
        df_test = deepcopy(df[df['subject#'] > 30])

        df_train, df_valid = split_df_on_column(df_train_valid, [1 - validation_size, validation_size], 'subject#')
        normalize_df_(df_train, other_dfs=[df_valid, df_test])
        df_res = get_split(df_train, df_valid, df_test, split)
        df_res.drop(columns='subject#', inplace=True)
        self.x, self.y = xy_split(df_res, y_columns)


class PowerPlant(RegressionDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00294/CCPP.zip'
        download_unzip(url, dataset_path)
        file_path = os.path.join(dataset_path, 'CCPP', 'Folds5x2_pp.xlsx')
        df = pd.read_excel(file_path)
        y_columns = ['PE']  # Not clear if this is the aim of the dataset
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class RealEstate(RegressionDataset):
    """
    Link to the dataset [description](http://archive.ics.uci.edu/ml/datasets/Real+estate+valuation+data+set).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'Real estate valuation data set.xlsx'
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00477/Real%20estate%20valuation%20data%20set.xlsx'
        download_file(url, dataset_path, filename)
        file_path = os.path.join(dataset_path, filename)

        df = pd.read_excel(file_path, index_col='No')
        y_columns = ['Y house price of unit area']
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class SensorLessDrive(ClassificationDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 11

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        file_name = 'data.csv'
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00325/Sensorless_drive_diagnosis.txt'
        download_file(url, dataset_path, file_name)
        file_path = os.path.join(dataset_path, file_name)
        df = pd.read_csv(file_path, header=None, sep=' ')
        y_columns = [48]
        label_encode_df_(df, y_columns[0])
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class Shuttle(ClassificationDataset):
    """
    Description of dataset [here](http://archive.ics.uci.edu/ml/datasets/Statlog+(Shuttle)).

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    num_classes = 7

    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url_train = 'http://archive.ics.uci.edu/ml/machine-learning-databases/statlog/shuttle/shuttle.trn.Z'
        url_test = 'http://archive.ics.uci.edu/ml/machine-learning-databases/statlog/shuttle/shuttle.tst'
        file_name_train = 'train.csv'
        file_name_test = 'test.csv'
        file_path_train = os.path.join(dataset_path, file_name_train)
        file_path_test = os.path.join(dataset_path, file_name_test)
        file_name_z = 'train.z'
        fresh_download = download_file(url_train, dataset_path, file_name_z)
        if fresh_download:
            path_z = os.path.join(dataset_path, file_name_z)
            with open(path_z, 'rb') as f_in:
                with open(file_path_train, 'wb') as f_out:
                    f_out.write(unlzw(f_in.read()))
            download_file(url_test, dataset_path, file_name_test)
        df_train_valid = pd.read_csv(file_path_train, header=None, sep=' ')
        y_columns = [9]
        df_train, df_valid = split_classification_df(df_train_valid,
                                                     [1 - validation_size, validation_size],
                                                     y_columns[0])
        df_test = pd.read_csv(file_path_test, header=None, sep=' ')
        label_encode_df_([df_train, df_valid, df_test], y_columns[0])
        normalize_df_(df_train, other_dfs=[df_valid, df_test], skip_column=y_columns[0])
        df_res = get_split(df_train, df_valid, df_test, split)
        self.x, self.y = xy_split(df_res, y_columns)
        self.y = self.y[:, 0]


class Superconductivity(RegressionDataset):
    """
    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00464/superconduct.zip'
        download_unzip(url, dataset_path)
        file_path = os.path.join(dataset_path, 'train.csv')
        df = pd.read_csv(file_path)
        y_columns = ['critical_temp']
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)


class WhiteWineQuality(RegressionDataset):
    """
    Description of dataset [here](http://archive.ics.uci.edu/ml/datasets/Wine+Quality).

    Citation:
    ```
    P. Cortez, A. Cerdeira, F. Almeida, T. Matos and J. Reis.
    Modeling wine preferences by data mining from physicochemical properties.
    In Decision Support Systems, Elsevier, 47(4):547-553, 2009.
    ```

    # Parameters
    root (str): Local path for storing/reading dataset files.
    split (str): One of {'train', 'validation', 'test'}
    validation_size (float): How large fraction in (0, 1) of the training partition to use for validation.
    """
    def __init__(self, root, split=TRAIN, validation_size=0.2):
        dataset_path = os.path.join(root, self.name)
        filename = 'data.csv'
        file_path = os.path.join(dataset_path, filename)
        url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-white.csv'
        download_file(url, dataset_path, filename)
        df = pd.read_csv(file_path, sep=';')
        y_columns = ['quality']
        self.x, self.y = split_normalize_sequence(df, y_columns, validation_size, split, self.type_)



import os
import zipfile
from copy import deepcopy
from urllib import request

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder, LabelEncoder

from ucimlr.constants import CLASSIFICATION
from ucimlr.constants import TRAIN, VALIDATION, TEST


class RandomStateContext:
    def __init__(self, seed=0):
        self.original_state = 0
        self.seed = seed

    def __enter__(self):
        self.original_state = np.random.get_state()
        np.random.seed(self.seed)

    def __exit__(self, type=None, value=None, traceback=None):
        np.random.set_state(self.original_state)


def download_file(url, dataset_path, filename):
    """
    Downloads the file at the url to the path 'dataset_path/filename'.

    Returns True if dataset was downloaded, False if already existing
    """
    os.makedirs(dataset_path, exist_ok=True)
    file_path = os.path.join(dataset_path, filename)
    if not os.path.exists(file_path):
        request.urlretrieve(url, file_path)
        return True
    else:
        return False


def download_unzip(url: str, dataset_path: str):
    """
    Downloads the file at the specified 'url' and unzips it
    at the specified 'dataset_path'.
    """
    filename_zip = 'data.zip'
    file_path_zip = os.path.join(dataset_path, filename_zip)
    fresh_download = download_file(url, dataset_path, filename_zip)
    if fresh_download:
        # TODO This naively assumes downloaded implies extracted
        zf = zipfile.ZipFile(file_path_zip)
        zf.extractall(dataset_path)


# Pandas DataFrame operations

def one_hot_encode_df_(df, skip_columns=None):
    """
    One-hot encodes all non-numeric columns and drops those columns.
    This is done in place.
    """
    if skip_columns is None:
        skip_columns = set()
    else:
        skip_columns = set(skip_columns)

    for col in set(df.columns) - skip_columns:
        if col in skip_columns:
            continue
        dtype = df[col].dtype
        if not np.issubdtype(dtype, np.number):
            x = df[col].to_numpy().reshape(-1, 1)
            one_hot = OneHotEncoder(sparse=False).fit_transform(x)
            for col_i in range(one_hot.shape[1]):
                df[f'{col}_{col_i}'] = one_hot[:, col_i]
            df.drop(columns=col, inplace=True)


def normalize_df_(df_train, *, other_dfs=None, skip_column=None):
    """
    Normalizes all columns of `df_train` to zero mean unit variance in place.
    Optionally performs same transformation to `other_dfs`

    # Parameters
    other_dfs [pd.DataFrame]: List of other DataFrames to apply transformation to
    skip_column (str, int): Column to omit, for example categorical targets.
    """
    if skip_column is None:
        skip_columns = set()
    else:
        skip_columns = {skip_column}

    # Skip where standard deviation is zero or close to zero
    low_std_columns = df_train.columns[df_train.std() < 1e-6]
    df_train.loc[:, low_std_columns] = 0
    if other_dfs is not None:
        for df in other_dfs:
            df.loc[:, low_std_columns] = 0
    skip_columns.update(set(low_std_columns))

    columns = list(set(df_train.columns) - skip_columns)
    mean = df_train[columns].mean(axis=0)
    std = df_train[columns].std(axis=0)
    df_train.loc[:, columns] -= mean
    df_train.loc[:, columns] /= std
    if other_dfs is not None:
        for df in other_dfs:
            df.loc[:, columns] -= mean
            df.loc[:, columns] /= std


def xy_split(df: pd.DataFrame, y_columns: list):
    """
    Takes a DataFrame and return X and Y numpy arrays where
    Y is given by the 'y_columns' argument.
    """
    x = df.drop(columns=y_columns).to_numpy()
    y = df[y_columns].to_numpy()
    return x, y


def label_encode_df_(dfs, y_column):
    """
    Label encodes in place.

    # Parameters
    dfs:  DataFrame or list of DataFrames
    y_column: The label column
    """
    if type(dfs) is not list:
        dfs = [dfs]
    ys = set()
    for df in dfs:
        ys.update(set(df[y_column]))
    le = LabelEncoder().fit(list(ys))
    for df in dfs:
        df[y_column] = le.transform(df[y_column])


def clean_na_(df):
    for column in df.columns:
        df[f'{column}_isnan'] = df[column].isna().astype(float)
    df.fillna(0.0, inplace=True)


def split_normalize_sequence(df: pd.DataFrame, y_columns, validation_size: float, split, dataset_type):
    """
    Performs the common sequence of operations:
    train_test_split -> normalize -> x, y split
    """
    if dataset_type == CLASSIFICATION:
        df_test, df_train, df_valid = split_classification_df(df,
                                                              [0.2, 0.8 - 0.8 * validation_size, 0.8 * validation_size],
                                                              y_columns[0])
    else:
        df_test, df_train, df_valid = split_df(df, [0.2, 0.8 - 0.8 * validation_size, 0.8 * validation_size])
    if dataset_type == CLASSIFICATION:
        normalize_df_(df_train, other_dfs=[df_valid, df_test], skip_column=y_columns[0])
    else:
        normalize_df_(df_train, other_dfs=[df_valid, df_test])
    df_res = get_split(df_train, df_valid, df_test, split)
    x, y = xy_split(df_res, y_columns)
    if dataset_type == CLASSIFICATION:
        y = y[:, 0]
    return x, y


def split_df(df: pd.DataFrame, fractions):
    """
    Randomly splits (always with same seed) the dataframe `df` into
    `fractions`.

    If you want one split to be always the same (given the same fraction of course),
    let that split be the first one. For example:
    ```
    df_test, df_train, df_valid = split_df(df, [0.2, 0.4, 0.4])
    ```
    or
    ```
    df_test, df_train, df_valid = split_df(df, [0.2, 0.7, 0.1])
    ```
    These always produce the same df_test.
    """
    fractions = np.array(fractions)
    if abs(1 - sum(fractions)) > 1e-6:
        raise ValueError('Fractions must sum to one.')
    for i in range(len(fractions)):
        fraction = fractions[0]
        sub_df = df.sample(frac=fraction, random_state=0)
        yield sub_df
        fractions = fractions[1:]
        if sum(fractions) > 0:
            fractions /= sum(fractions)
        df = df.drop(sub_df.index)


def split_classification_df(df, fractions, y_column):
    """
    Like split_df but groups into respective classes and then samples
    fractions of those groups. Makes sure that:
    1) No labels/classes are missing from any of the splits
    2) The proportions of classes are the same in all splits
    """
    fractions = np.array(fractions)
    splits = [pd.DataFrame(columns=df.columns)] * len(fractions)
    for i in range(len(fractions)):
        fraction = fractions[0]
        for value, group in df.groupby(y_column):
            sample = group.sample(frac=fraction, random_state=0)
            splits[i] = splits[i].append(sample)
            df = df.drop(sample.index)
        fractions = fractions[1:]
        if fractions.sum() > 0:
            fractions /= fractions.sum()
    # np.isnan crashes if this cast is not done
    for i, split in enumerate(splits):
        splits[i] = split.astype(np.float32)
        splits[i].loc[:, y_column] = split.loc[:, y_column].astype(np.int)
    return splits


def split_df_on_column(df, fractions, column):
    """
    Randomly splits (always with same seed) the dataframe `df` into
    `fractions`. Values in the `column` are disjoint over the splits.
    """
    fractions = np.array(fractions)
    if sum(fractions) != 1.0:
        raise ValueError('Fractions must sum to one.')
    with RandomStateContext():
        groups = np.random.permutation(df[column].unique())
    for i in range(len(fractions)):
        fraction = fractions[0]
        groups_sample = groups[:int(fraction * len(groups))]
        df_sample = deepcopy(df.loc[df[column].isin(groups_sample), :])
        yield df_sample
        fractions = fractions[1:]
        if sum(fractions) > 0:
            fractions /= sum(fractions)
        df = df.drop(df_sample.index)


def get_split(df_train, df_valid, df_test, split):
    if split == TRAIN:
        return df_train
    elif split == VALIDATION:
        return df_valid
    elif split == TEST:
        return df_test
    else:
        raise ValueError('split is not correct, see ucimlr.constants')
